from sqlalchemy import BigInteger, String, select
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    Session,
    mapped_column
)

from ..constants import InvokerEnvironment


class Base(DeclarativeBase):
    pass


class DeployedModel(Base):
    __tablename__ = "deployed_models"
    env: Mapped[InvokerEnvironment] = mapped_column(
        String(10), primary_key=True
    )
    model_id: Mapped[int] = mapped_column(BigInteger(), primary_key=True)
    image_id: Mapped[str] = mapped_column(String(71))
    model_name: Mapped[str] = mapped_column(String(50))

    def __repr__(self) -> str:
        return f"DeployedModel(env={self.env!r}, \
            model_id={self.model_id!r}, model_name={self.model_name!r})"


class InferenceLookup(Base):
    __tablename__ = "inference_lookup"
    env: Mapped[InvokerEnvironment] = mapped_column(
        String(10), primary_key=True
    )
    model_id: Mapped[int] = mapped_column(BigInteger(), primary_key=True)
    inference_id: Mapped[str] = mapped_column(BigInteger(), primary_key=True)

    def __repr__(self) -> str:
        return f"InferenceModel(env={self.env!r}, \
            model_id={self.model_id!r}, inference_id={self.inference_id!r})"


def addDeployedModel(
    engine,
    env: InvokerEnvironment,
    model_id: int,
    image_id: str,
    model_name: int,
):
    with Session(engine) as session:
        obj = DeployedModel(
            env=env,
            model_id=model_id,
            image_id=image_id,
            model_name=model_name,
        )
        session.add(obj)
        session.commit()


def addInference(
    engine, env: InvokerEnvironment, model_id: int, inference_id: int
) -> bool:
    with Session(engine) as session:
        obj = InferenceLookup(
            env=env, model_id=model_id, inference_id=inference_id
        )
        session.add(obj)
        session.commit()
    return True


def checkInference(
    engine, env: InvokerEnvironment, model_id: int, inference_id: int
) -> bool:
    with Session(engine) as session:
        stmt = (
            select(InferenceLookup)
            .where(InferenceLookup.env == env)
            .where(InferenceLookup.model_id == model_id)
            .where(InferenceLookup.inference_id == inference_id)
        )
        inf = session.scalars(stmt).one_or_none()
        if inf:
            return True
    return False


def getDeployedModels(engine) -> list[DeployedModel]:
    with Session(engine) as session:
        stmt = select(DeployedModel)
        result = []
        for model in session.scalars(stmt):
            result.append(model)
        return result


def getImageIdByModel(engine, env: InvokerEnvironment, model_id: int) -> str:
    with Session(engine) as session:
        stmt = (
            select(DeployedModel)
            .where(DeployedModel.model_id == model_id)
            .where(DeployedModel.env == env)
        )
        model = session.scalars(stmt).one()
        return model.image_id


def updateModel(engine, model_id: int, image_id: str):
    with Session(engine) as session:
        stmt = select(DeployedModel).where(DeployedModel.model_id == model_id)
        model = session.scalars(stmt).one()
        model.image_id = image_id
        session.commit()



import json
import logging
import os
import docker
from ..ast_check import model_check


def cmd_build(args, config):
    filesTocCheck = ["Dockerfile"]
    client = docker.DockerClient(base_url=config["system"]["docker_url"])
    for f in filesTocCheck:
        if not os.path.isfile(f):
            logging.info("Mising file {file}".format(file=f))
            os._exit(1)
    logging.info("Building...")
    try:
        dockerfile = os.getcwd()
        print(dockerfile)
        (img, logs) = client.images.build(path=dockerfile)
        print(logs)
        for log in logs:
            print(log)
        build = {"id": str(img.id)}
        logging.info("Image ID {}".format(img.id))
        f = open("./build.json", "w")
        f.write(json.dumps(build))
        f.close()

        if not os.path.isfile("./model.py"):
            logging.info("Missing file {file}".format(file="model.py"))
            os._exit(1)

        model_check()

    except Exception as e:
        print("exception here... handle later")
        print(e)



import inspect

from .inputbase import InvokerNetworkInputBase


def invoker(name: str = None, tags: list[str] = [], description: str = None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            print("pre invoker wapper")
            res = func(*args, **kwargs)
            print("post invoker wrapper")
            return res

        return wrapper

    return decorator


def invoker_input(cls):
    def wrap(cls):
        return _process_class(cls)

    # See if we're being called as @dataclass or @dataclass().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    return wrap(cls)


def _process_class(cls):
    cls_annotations = cls.__dict__.get("__annotations__", {})
    annotations = cls_annotations.items()
    for name, type in annotations:
        # all annotatioins must inherit from InvokerNetworkInputBase
        if not issubclass(type, InvokerNetworkInputBase):
            raise TypeError(
                "Attribute '{attribute}' having {child} type \
                    needs to inherit from {parent}. Please use \
                        Invoker Network Input types only".format(
                    attribute=name,
                    child=type.__name__,
                    parent=InvokerNetworkInputBase.__name__,
                )
            )
        member = getattr(cls, name, None)
        if member is None:
            print(type)
            print("member is none init please {}, {}".format(name, type))
            setattr(cls, name, type())

    attributes = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a)))
    members = list(
        filter(
            (lambda a: not (a[0].startswith("__") and a[0].endswith("__"))),
            attributes,
        )
    )
    for name, obj in members:
        if not issubclass(obj.__class__, InvokerNetworkInputBase):
            raise ValueError(
                "Expecting Invoker Network Input type for \
                    '{attribute}' but got {wrongtype}".format(
                    attribute=name, wrongtype=obj.__class__.__name__
                )
            )
        obj.setName(name)

    def toJSON(self):
        attributes = inspect.getmembers(
            self, lambda a: not (inspect.isroutine(a))
        )
        members = list(
            filter(
                lambda a: not (a[0].startswith("__") and a[0].endswith("__")),
                attributes,
            )
        )
        ret = {}
        ret["inputs"] = []
        for name, obj in members:
            ret["inputs"].append(obj.__dict__)
        return ret

    # inject toSON
    cls.toJSON = toJSON

    return cls



import io
import zipfile


class InvokerOutput:
    outputs = []
    fileCounter = 0

    def addText(content: str, filename=None):
        if filename:
            InvokerOutput.outputs.append(
                (filename, io.BytesIO(content.encode("utf-8")))
            )
        else:
            InvokerOutput.outputs.append(
                (
                    str(InvokerOutput.fileCounter),
                    io.BytesIO(content.encode("utf-8")),
                )
            )
            InvokerOutput.fileCounter += 1

    def done():
        zip_buffer = io.BytesIO()
        print(InvokerOutput.outputs)
        with zipfile.ZipFile(
            zip_buffer, "a", zipfile.ZIP_DEFLATED, False
        ) as zip_file:
            for file_name, data in InvokerOutput.outputs:
                zip_file.writestr(
                    "output/{}".format(file_name), data.getvalue()
                )

        with open("/invoker/output.zip", "wb") as f:
            f.write(zip_buffer.getvalue())



import logging

import aiohttp
import requests


def status_ipfs(url) -> bool:
    res = requests.get(url)
    return res.status_code == 200


def send_to_ipfs(input: bytes, url: str) -> str:
    logging.info("ipfs post {}".format(url))
    res = requests.post(
        url=url,
        data=input,
        headers={"Content-Type": "application/octet-stream"},
    )
    js = res.json()
    return js["cid"]


def get_from_ipfs(hash: str, url: str) -> bytes:
    urlToGet = "{}/{}".format(url, hash)
    logging.info("ipfs url get {}".format(urlToGet))
    res = requests.get(urlToGet)
    return res.content


async def async_get_from_ipfs(loop, hash: str, url: str) -> bytes:
    async with aiohttp.ClientSession(loop=loop) as session:
        urlToGet = "{}/{}".format(url, hash)
        async with session.get(urlToGet) as response:
            output = await response.read()
            return output


async def async_send_to_ipfs(loop, input: bytes, url: str) -> str:
    async with aiohttp.ClientSession(loop=loop) as session:
        async with session.post(
            url,
            data=input,
            headers={"Content-Type": "application/octet-stream"},
        ) as response:
            output = await response.json()
            return output["cid"]



from .inputbase import InvokerNetworkInputBase


class InvokerNetworkSelectInput(InvokerNetworkInputBase):
    def __init__(
        self,
        description: str = None,
        required: bool = False,
        options: list[str] = [],
    ):
        super().__init__(description, required)
        if len(options) == 0:
            raise ValueError("Options cannot be empty")
        self.type = "select"
        self.options = options



from sqlalchemy import create_engine

from .deployed_models import Base


def get_engine(db_path):
    return create_engine("sqlite:///{}".format(db_path))


def init_tables(engine):
    Base.metadata.create_all(engine)



from .inputbase import InvokerNetworkInputBase


class InvokerNetworkFileInput(InvokerNetworkInputBase):
    def __init__(self, description: str = None, required: bool = False):
        super().__init__(description, required)
        self.type = "file"



import logging
import os

from ..cores.wallet import generatewallet

heart_beat_url = "https://hb.invoker.network/app"

ipfs_post_gateway = "https://ipfs.invoker.network/app/ipfs-route"
ipfs_get_gateway = "https://ipfs.invoker.network/gateway/ipfs/"
ipfs_test_hash = "QmdUUfATjEKuhNQacktJJWUydfz7Zh3xfwKAkRfXU7eNB9"

docker_url = "unix:///var/run/docker.sock"
docker_sock = "tcp://127.0.0.1:1234"
xml_rpc_addr = "localhost"
xml_rpc_port = 9005
tmp_folder = "/tmp/"

local_rpc_url = "http://127.0.0.1:8899"
local_websocket_url = "ws://localhost:8900"

dev_rpc_url = "https://api.devnet.solana.com"
dev_websocket_url = (
    "wss://multi-proportionate-firefly.solana-"
    "devnet.discover.quiknode.pro/8149116fe82f66981c679c6b701e03a67fb1cb68/"
)

mainnet_rpc_url = "https://api.devnet.solana.com"
mainnet_websocket_url = (
    "wss://multi-proportionate-firefly.solana-devnet"
    ".discover.quiknode.pro/8149116fe82f66981c679c6b701e03a67fb1cb68/"
)


def cmd_init(args, config):
    logging.info("Initializing config...")
    filepath = args.conf[-1]
    if os.path.isfile(filepath):
        logging.info("Config file exists. I can't overwrite it")
        os._exit(-1)
    else:
        p = os.path.dirname(os.path.abspath(filepath))
        logging.info("Directory is {}".format(p))
        os.makedirs(p, exist_ok=True)
    config["system"] = {}
    config["system"]["heart_beat_url"] = str(heart_beat_url)
    config["system"]["ipfs_post_gateway"] = str(ipfs_post_gateway)
    config["system"]["ipfs_get_gateway"] = str(ipfs_get_gateway)
    config["system"]["ipfs_test_hash"] = str(ipfs_test_hash)
    config["system"]["docker_url"] = str(docker_url)
    config["system"]["docker_sock"] = str(docker_sock)
    config["system"]["xml_rpc_addr"] = str(xml_rpc_addr)
    config["system"]["xml_rpc_port"] = str(xml_rpc_port)
    config["system"]["tmp_folder"] = str(tmp_folder)
    config["system"]["log_level"] = str(20)
    config["system"]["wallet"] = generatewallet().to_json()
    config["local"] = {}
    config["local"]["rpc_url"] = local_rpc_url
    config["local"]["websocket_url"] = local_websocket_url
    config["dev"] = {}
    config["dev"]["rpc_url"] = dev_rpc_url
    config["dev"]["websocket_url"] = dev_websocket_url
    config["mainnet"] = {}
    config["mainnet"]["rpc_url"] = mainnet_rpc_url
    config["mainnet"]["websocket_url"] = mainnet_websocket_url

    configfile = open(filepath, "w")
    logging.info("saving to file {}".format(filepath))
    config.write(configfile)



from .inputbase import InvokerNetworkInputBase


class InvokerNetworkRangeInput(InvokerNetworkInputBase):
    def __init__(
        self,
        description: str = None,
        required: bool = False,
        min: int = None,
        max: int = None,
        step: int = None,
    ):
        super().__init__(description, required)
        self.type = "range"
        if min is None:
            self.min = -1
        if max is None:
            self.max = -1
        if step is None:
            self.step = -1
        self.min = min
        self.max = max
        self.step = step



import json


class InvokerNetworkInputBase:
    def __init__(self, description: str = None, required: bool = False):
        self.name = None
        if description is None:
            self.description = ""
        else:
            self.description = description
        self.required = required

    def toJSON(self):
        return json.dumps(
            self, default=lambda o: o.__dict__, sort_keys=True, indent=4
        )

    def setName(self, name):
        self.name = name



import logging
import xmlrpc.client


def connect_daemon(host: str, port: str):
    return xmlrpc.client.ServerProxy(
        "http://{host}:{port}".format(host=host, port=port)
    )


def get_client(host: str, port: str):
    try:
        s = connect_daemon(host, port)
        if s.isAlive() is not True:
            return None
        return s
    except ConnectionRefusedError as err:
        logging.debug(f"Connection refused {err}")
        return None
    except Exception as err:
        logging.debug("we got exception here")
        print(f"Unexpected {err=}, {type(err)=}")
        return None



from .inputbase import InvokerNetworkInputBase


class InvokerNetworkTextInput(InvokerNetworkInputBase):
    def __init__(
        self,
        description: str = None,
        required: bool = False,
        minChar: int = None,
        maxChar: int = None,
        textarea: bool = False,
    ):
        super().__init__(description, required)
        if minChar is None:
            self.minchar = -1
        else:
            if not isinstance(minChar, int) or minChar < 0:
                raise ValueError("Invalid Value for {}".format("min char"))
            self.minchar = minChar
        if maxChar is None:
            self.maxchar = -1
        else:
            if not isinstance(maxChar, int) or maxChar < 0:
                raise ValueError("Invalid Value for {}".format("max char"))
            self.maxchar = maxChar
        self.textarea = textarea
        self.type = "text"



from invoker_terminal import (
    InvokerNetworkFileInput,
    InvokerNetworkNumberInput,
    InvokerNetworkRangeInput,
    InvokerNetworkSelectInput,
    InvokerNetworkTextInput,
    InvokerOutput,
    invoker,
    invoker_input,
)


@invoker_input
class AsampleClass:
    count = InvokerNetworkNumberInput(
        description="count variable used in the model",
        required=True,
        minNum=100,
        maxNum=1000,
    )

    selectedValue = InvokerNetworkSelectInput(
        description="Important selections need to be done",
        required=True,
        options=["options1", "options2", "options3", "options4"],
    )

    count2 = InvokerNetworkNumberInput(
        description="another variable but required",
        required=True,
        minNum=100,
        maxNum=1000,
    )
    prompt = InvokerNetworkTextInput(
        description="my description here",
        required=True,
        minChar=10,
        maxChar=12,
        textarea=False,
    )
    prompt2 = InvokerNetworkTextInput(
        description="Prompt to execute in the model",
        required=True,
        textarea=True,
    )

    sliderval = InvokerNetworkRangeInput(
        description="Slider value", required=True, min=0, max=100, step=25
    )

    profilepicture = InvokerNetworkFileInput(
        description="Your profile picture", required=True
    )


@invoker(
    name="Web LLM",
    tags=["LLM", "15B Params", "4-bit Quantization"],
    description="This project brings language model chats \
        directly onto web browsers. Everything runs inside \
            the browser with no server support and accelerated with WebGPU. \
                We can bring a lot of fun opportunities to \
                    build AI assistants for everyone \
                        and enable privacy while enjoying GPU acceleration.",
)
def invoke(input: AsampleClass):
    print("I got invoked")
    newCount = input.count * 200
    newPrompt = ""
    newPrompt += input.prompt
    newPrompt += input.prompt2
    InvokerOutput.addText("new count is {}".format(newCount))
    InvokerOutput.addText("new Prompt is {}".format(newPrompt))
    InvokerOutput.addText("my content is here", "5.txt")



import asyncio
import atexit
import json
import os
import threading
import time
import traceback
import uuid
from multiprocessing import current_process
from pathlib import Path
from struct import pack_into
from xmlrpc.server import (
    SimpleXMLRPCRequestHandler,
    SimpleXMLRPCServer,
)

import aiodocker
import aiohttp
import websockets
from nacl.encoding import Base64Encoder
from nacl.signing import SigningKey
from solana.rpc.websocket_api import connect
from solders.pubkey import Pubkey

from .cores.anchor import descrow, getEscrow, getModel
from .cores.ipfs import async_get_from_ipfs, async_send_to_ipfs
from .cores.wallet import getPubkey, loadwallet
from .storage.deployed_models import (
    DeployedModel,
    checkInference,
    getDeployedModels,
    getImageIdByModel,
)
from .storage.helper import get_engine
from .utils import getTaskId

import logging
from logging.handlers import RotatingFileHandler


idl = open(
    os.path.dirname(__file__) + "/idl/invoker_network_market.json"
).read()
idl = json.loads(idl)


def exit_handler():
    print("Daemon quitting..do some cleanup")


# Restrict to a particular path.
class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/RPC2",)


logger = None

quitFlag = 0


def quit():
    global quitFlag
    quitFlag = 1
    logger.info("quiting....")
    return 1


def isAlive():
    return True


tasks_listening = {}


async def docker_run(cf, input: bytes, image_id: str) -> bytes:
    logger = logging.getLogger("DockerExecutor")
    id = uuid.uuid4()
    host_url = os.path.join(cf["system"]["tmp_folder"], str(id))
    Path(host_url).mkdir(parents=True, exist_ok=True)
    logger.info("Host url {}".format(host_url))
    with open("{}/input.zip".format(host_url), "wb") as f:
        f.write(input)

    docker = aiodocker.Docker(cf["system"]["docker_url"])
    container_config = {
        "Image": "{}".format(image_id),
        "HostConfig": {"Binds": [f"{host_url}:/invoker"]},
    }
    logger.info("Container ID {}".format(str(id)))
    logger.info("Container Config {}".format(container_config))
    container = await docker.containers.create_or_replace(
        str(id), container_config
    )
    await container.start()
    await container.wait()
    await container.delete(force=True)
    await docker.close()
    output = open("{}/output.zip".format(host_url), "rb").read()
    # todo
    # delete host mount point here
    return output


async def heart_beat(cf, model: DeployedModel):
    HEARTBEAT_INTERVAL = 15
    logger = logging.getLogger("HeartBeat")
    keystr = cf["system"]["wallet"]
    keypair = loadwallet(keystr)
    sk = SigningKey(keypair.secret())
    url = cf["system"]["heart_beat_url"]
    modelId = model.model_id
    env = model.env

    while quitFlag != 1:
        status = tasks_listening.get("{}_{}".format(env, modelId), False)
        if env == "local" or env == "dev":
            logger.info("Heartbeat is disabled for local and dev environments")
            status = False

        if status:
            logger.info(
                "pumping every {} sec for {} - [status] {}".format(
                    HEARTBEAT_INTERVAL, modelId, status
                )
            )
            msg = {}
            msg["ts"] = time.time()
            msg["pk"] = str(keypair.pubkey())
            msg["modelid"] = modelId
            msg["content"] = "hb"
            jsonstr = json.dumps(msg)
            signed = sk.sign(jsonstr.encode("utf-8"), encoder=Base64Encoder)
            postpayload = {}
            postpayload["modelid"] = modelId
            postpayload["pk"] = str(keypair.pubkey())
            postpayload["sig"] = str(signed)
            async with aiohttp.ClientSession() as session:
                try:
                    async with session.post(url, json=postpayload) as resp:
                        print(resp.status)
                        # print(await resp.text())
                        logger.info("OK - {}".format(modelId))
                except Exception as e:
                    logger.error("FAIL - {}".format(modelId))
                    logger.error(e)
                finally:
                    await session.close()

        await asyncio.sleep(HEARTBEAT_INTERVAL)
    logger.info("is quiting...")


async def invoke_executor(args, cf, loop, engine, env, modelid, inferenceid):
    logger = logging.getLogger(
        "InvokeExecutor for {} {} {}".format(env, modelid, inferenceid)
    )
    try:
        escrowAccount = await getEscrow(cf, modelid, inferenceid)
        ipfsfile = await async_get_from_ipfs(
            loop, escrowAccount.input, cf["system"]["ipfs_get_gateway"]
        )
        imageid = getImageIdByModel(engine, env, modelid)
        logger.info("Fetch image id {}".format(imageid))
        output = await docker_run(cf, ipfsfile, imageid)
        ipfsoutput = await async_send_to_ipfs(
            loop, output, cf["system"]["ipfs_post_gateway"]
        )
        logger.info("Output hash {}".format(ipfsoutput))
        await descrow(cf, modelid, inferenceid, ipfsoutput)
    except Exception as e:
        logger.info("{} {} {} {} failed".format(env, modelid, inferenceid, e))


async def invoke_processor(args, cf, invokeQ: asyncio.Queue):
    logger = logging.getLogger("Invoke Processor")
    logger.info("starting...")
    engine = get_engine(args.db_path)
    while quitFlag != 1:
        (env, modelid, inferenceid) = await invokeQ.get()
        loop = asyncio.get_event_loop()
        loop.create_task(
            invoke_executor(args, cf, loop, engine, env, modelid, inferenceid)
        )
    # parse the inputs
    # execute in docker
    # if everything is okay
    # descrow
    # if evrything is okay
    # mark the database as task is done.
    logger.info("done...")
    pass


async def notify_processor(
    args, config, engine, notifyQ: asyncio.Queue, invokeQ: asyncio.Queue
):
    logger = logging.getLogger("Notify Processor")
    try:
        while quitFlag != 1:
            logger.info("waiting for processing...")
            (env, modelid) = await notifyQ.get()
            model = await getModel(config, modelid)
            last_inference = model.inference_count - 1
            if not checkInference(engine, env, modelid, last_inference):
                # invokeQ()
                logger.info(
                    "inference not found {} {} {}".format(
                        env, modelid, last_inference
                    )
                )
                await invokeQ.put((env, modelid, last_inference))
                pass
            # get account info for the model
            # get latest inference
            # check if it is in our storage marked as success.
            # if not send to the invoke que
            logger.info("item {} on {} has been \
                processed".format(env, modelid))
    except EOFError:
        logger.info("Connection closed brutally")
    logger.info("is quitting...")


tasks = {}


async def listen_account_change(cf, notifyQ, model: DeployedModel):
    logger = logging.getLogger("Connection")
    env = model.env
    wss = cf[env]["websocket_url"]
    buffer = bytearray(16)
    modelId = model.model_id
    pack_into("<I", buffer, 0, modelId)
    modelAddr, _ = Pubkey.find_program_address(
        seeds=[bytes(buffer), bytes("model", "utf-8")],
        program_id=getPubkey(idl["metadata"]["address"]),
    )
    logger.info(
        "starting to listening model account changes for {} on {}".format(
            modelId, env
        )
    )
    subscription_id = None
    try:
        async with connect(wss) as websocket:
            await websocket.account_subscribe(modelAddr, encoding="base64")
            logger.info("I subscribe websocket")
            first_resp = await websocket.recv()
            # print(first_resp)
            logger.info("first resp")
            logger.info(first_resp)
            subscription_id = first_resp[0].id
            # subscription_id = first_resp.result
            tasks_listening["{}_{}".format(env, modelId)] = True
            async for msg in websocket:
                logger.info(
                    "Account has been changed for {} {}".format(env, modelId)
                )
                await notifyQ.put((env, modelId))
    except websockets.exceptions.ConnectionClosedError:
        logger.info("Connection closed... But trying to reconnect again")
        await asyncio.sleep(3)
        pass
    except asyncio.exceptions.IncompleteReadError:
        logger.info("Incomplete read error...***")
        await asyncio.sleep(3)
        pass
    except OSError:
        logger.info("Os Error here but keep moving")
        await asyncio.sleep(3)
        pass
    finally:
        tasks_listening["{}_{}".format(env, modelId)] = False
        tb = traceback.format_exc()
        logger.info(tb)
        try:
            logger.info("trying to unsub here for {}".format(subscription_id))
            if subscription_id:
                logger.info("really unsubbing here")
                # await websocket.logs_unsubscribe(subscription_id)
        except asyncio.exceptions.IncompleteReadError:
            logger.info("it is fine keep going")
            pass


def get_contracts_list():
    dummy = {}
    for key in tasks_listening.keys():
        dummy[str(key)] = tasks_listening[key]
    return dummy


async def start_account_listener(cf, notifyQ, model: DeployedModel):
    task_id = getTaskId(model)
    logger = logging.getLogger("Connection")
    while True:
        if quitFlag == 1:
            break

        if task_id in tasks and not tasks[task_id].done():
            logger.info(f"Task {task_id} is already running")
        else:
            # Create a new event loop for the task
            loop = asyncio.get_running_loop()

            # Start a new task on the new event loop
            tasks[task_id] = loop.create_task(
                listen_account_change(cf, notifyQ, model)
            )
            logger.info(f"Started new task {task_id}")

        # Wait for the task to complete before running it again
        await tasks[task_id]


async def kill_switch(cf):
    logger = logging.getLogger("Daemon")
    while True:
        if quitFlag == 1:
            loop = asyncio.get_event_loop()
            loop.stop()
            break
        logger.info("kill switch quit flag {}".format(quitFlag))
        await asyncio.sleep(5)


async def start_tasks(args, cf):
    notifyQ = asyncio.Queue()
    invokeQ = asyncio.Queue()
    engine = get_engine(args.db_path)
    models = getDeployedModels(engine)
    tasklist = [
        kill_switch(cf),
        notify_processor(args, cf, engine, notifyQ, invokeQ),
        invoke_processor(args, cf, invokeQ),
    ]
    for model in models:
        tasklist.append(start_account_listener(cf, notifyQ, model))
        tasklist.append(heart_beat(cf, model))
    await asyncio.gather(*tasklist)


def account_listener(args, cf):
    logger = logging.getLogger("Connection")
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        asyncio.run(start_tasks(args, cf))
    except RuntimeError:
        logger.info("Runtime error received from asyncio loop")
        pass
    logger.info("account listener is quitting...")


# function to be executed in a new process
def task(argz, config):
    global quitFlag
    global logger
    ###
    logrotate = RotatingFileHandler(
        filename=argz.log_path,
        mode="a",
        maxBytes=5 * 1024 * 1024,
        backupCount=1,
        encoding=None,
        delay=0,
    )

    logging.basicConfig(
        format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
        level=int(config["system"]["log_level"]),
        handlers=[logrotate],
    )
    logger = logging.getLogger("Daemon")

    logging.info("from daemon {}".format(argz))
    logging.info("from daemon {}".format(config))
    # get the current process
    process = current_process()
    atexit.register(exit_handler)
    # report if daemon process
    logging.info(f"Daemon process: {process.daemon}")
    host = config["system"]["xml_rpc_addr"]
    port = config["system"]["xml_rpc_port"]
    server = SimpleXMLRPCServer(
        (host, int(port)), requestHandler=RequestHandler, logRequests=False
    )
    server.register_function(quit)
    server.register_function(isAlive)
    server.register_function(get_contracts_list)

    threads = []
    threads.append(
        threading.Thread(target=account_listener, args=(argz, config))
    )

    for t in threads:
        t.start()

    while quitFlag != 1:
        server.handle_request()

    loop = asyncio.get_event_loop()
    loop.close()

    for t in threads:
        t.join()



import ast
import importlib
import json
import logging
import os
import sys
from typing import List

INVOKER_DECORATOR = "invoker"
INVOKER_INPUT = "invoker_input"
INVOKER_NETWORK_INPUT = "InvokerNetworkInput"
INVOKER_INPUT_NAME = "name"
INVOKER_INPUT_DESC = "description"
ALLOWED_KEYWORDS = [INVOKER_INPUT_NAME, INVOKER_INPUT_DESC]


def model_check():
    logging.info("checking model inputs/outputs")
    f = open("./model.py", "r")
    root = ast.parse(f.read(), "model.py")
    f.close()
    # only one function with @invoker decorator
    invokerDecoratorCount = 0
    invokerFunction: ast.FunctionDef = None
    invokerInputDecoratorCount = 0
    invokerInputClass: ast.ClassDef = None
    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.FunctionDef):
            decorators = node.decorator_list
            for decorator in decorators:
                if isinstance(decorator, ast.Call):
                    if decorator.func.id == INVOKER_DECORATOR:
                        invokerDecoratorCount += 1
                        invokerFunction = node
                if (
                    isinstance(decorator, ast.Name)
                    and decorator.id == INVOKER_DECORATOR
                ):
                    invokerDecoratorCount += 1
                    invokerFunction = node

        if isinstance(node, ast.ClassDef):
            decorators = node.decorator_list
            for decorator in decorators:
                if decorator.id == INVOKER_INPUT:
                    invokerInputDecoratorCount += 1
                    invokerInputClass = node

    if not invokerInputDecoratorCount == 1:
        logging.info(
            "Model needs to have @invoker_input decorator to an input class"
        )
        os._exit(1)

    if not invokerDecoratorCount == 1:
        logging.info("Model can have only one @invoker decorator")
        os._exit(1)

    arguments: ast.arguments = invokerFunction.args
    args: List[ast.arg] = arguments.args
    if not len(args) == 1:
        logging.info("Invoker function should have only one input")
        os._exit(1)

    # make sure input passed to the invoker function
    # is actually invokerInputClass
    desc = {}
    print(ast.dump(invokerInputClass))
    spec = importlib.util.spec_from_file_location(
        "model", os.path.abspath("./model.py")
    )
    foo = importlib.util.module_from_spec(spec)
    sys.modules["model"] = foo
    spec.loader.exec_module(foo)
    cls = getattr(foo, invokerInputClass.name)
    obj = cls()
    inputs = obj.toJSON()
    desc.update(inputs)
    print(ast.dump(invokerFunction))
    decorator_list = invokerFunction.decorator_list
    for decorator in decorator_list:
        if isinstance(decorator, ast.Call):
            keywords = decorator.keywords
            for keyword in keywords:
                if keyword.arg in ALLOWED_KEYWORDS:
                    val: ast.Constant = keyword.value
                    desc[keyword.arg] = val.value
    jsonformatted = json.dumps(desc, sort_keys=True, indent=4)
    f = open("./desc.json", "w")
    f.write(jsonformatted)
    f.close()
    print(jsonformatted)
    logging.info("Success")



from ..storage.deployed_models import checkInference
from ..storage.helper import get_engine


def cmd_ipfs(args, conf):
    print("args ")
    print(args)
    print("conf")
    print(conf)
    engine = get_engine(args.db_path)
    # res = addInference(engine, "local", 1, 1)
    # print(res)
    res2 = checkInference(engine, "local", 1, 2)
    print(res2)



import importlib
import os
import sys

from ..outputs.output import InvokerOutput
from ..cores.invoker_init import ParseInput

def cmd_invoke(config, args):
    spec = importlib.util.spec_from_file_location(
        "model", os.path.abspath("./model.py")
    )
    foo = importlib.util.module_from_spec(spec)
    sys.modules["model"] = foo
    spec.loader.exec_module(foo)
    fn = getattr(foo, "invoke")
    print(fn)
    dummy = ParseInput()
    run = fn(dummy)
    print(run)
    InvokerOutput.done()



import json
import os
from struct import pack
from typing import Tuple

from anchorpy import Context, Idl, Program, Provider, Wallet
from solana.rpc.api import Client
from solana.rpc.async_api import AsyncClient
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.client import Token
from spl.token.constants import TOKEN_PROGRAM_ID
from spl.token.core import AccountInfo
from spl.token.instructions import get_associated_token_address
from .wallet import getPubkey, loadwallet

max_int64 = 0xFFFFFFFFFFFFFFFF


class InvokerModel:
    owner: Pubkey
    description: str
    price: int
    token_mint: Pubkey
    inference_count: int


class EscrowAccount:
    requester: Pubkey
    input: str
    output: str


rawidl = open(
    os.path.dirname(__file__) + "/../idl/invoker_network_market.json"
).read()
# rawidl = json.load(rawidl)
idlparsed = json.loads(rawidl)


def getMarketSettingsPDA() -> Tuple[Pubkey, int]:
    marketSettingsPDA, marketSettingsBump = Pubkey.find_program_address(
        seeds=[bytes("market-settings", "utf-8")],
        program_id=getPubkey(idlparsed["metadata"]["address"]),
    )

    return marketSettingsPDA, marketSettingsBump


def getEscrowWalletPDA(modelId: int, inferenceid: int) -> Pubkey:
    # inferencebuffer = bytearray(16)
    # modelbuffer = bytearray(16)
    # pack_into("<I", modelbuffer, 0, modelId)
    # pack_into("<I", inferencebuffer, 0, inferenceid)
    inferencebuffer = pack(
        "<QQ", inferenceid & max_int64, (inferenceid >> 64) & max_int64
    )
    modelbuffer = pack("<QQ", modelId & max_int64, (modelId >> 64) & max_int64)

    escrowWalletPDA, _ = Pubkey.find_program_address(
        seeds=[
            bytes(inferencebuffer),
            bytes(modelbuffer),
            bytes("escrow-wallet", "utf-8"),
        ],
        program_id=getPubkey(idlparsed["metadata"]["address"]),
    )
    return escrowWalletPDA


def getEscrowAccountPDA(modelId: int, inferenceid: int) -> Pubkey:
    # inferencebuffer = bytearray(16)
    # modelbuffer = bytearray(16)
    # pack_into("<I", modelbuffer, 0, modelId)
    # pack_into("<I", inferencebuffer, 0, inferenceid)
    inferencebuffer = pack(
        "<QQ", inferenceid & max_int64, (inferenceid >> 64) & max_int64
    )
    modelbuffer = pack("<QQ", modelId & max_int64, (modelId >> 64) & max_int64)

    escrowPDA, _ = Pubkey.find_program_address(
        seeds=[
            bytes(inferencebuffer),
            bytes(modelbuffer),
            bytes("escrow", "utf-8"),
        ],
        program_id=getPubkey(idlparsed["metadata"]["address"]),
    )
    return escrowPDA


def getModelPDA(modelId: int) -> Pubkey:
    max_int64 = 0xFFFFFFFFFFFFFFFF
    packed = pack("<QQ", modelId & max_int64, (modelId >> 64) & max_int64)
    modelPDA, _ = Pubkey.find_program_address(
        seeds=[bytes(packed), bytes("model", "utf-8")],
        program_id=getPubkey(idlparsed["metadata"]["address"]),
    )
    return modelPDA


def getWallet(config):
    keystr = config["system"]["wallet"]
    keypair = loadwallet(keystr)
    return Wallet(keypair)


def getKeypair(config):
    keystr = config["system"]["wallet"]
    keypair = loadwallet(keystr)
    return keypair


def getProgram(config) -> Program:
    wallet = getWallet(config)
    client = AsyncClient(config["active"]["rpc_url"])
    provider = Provider(client, wallet)
    program_id = getPubkey(idlparsed["metadata"]["address"])
    idlAnchor = Idl.from_json(rawidl)
    return Program(idlAnchor, program_id, provider)


async def getEscrow(config, modelId: int, inferenceId: int) -> EscrowAccount:
    async with getProgram(config) as program:
        escrowPDA = getEscrowAccountPDA(modelId, inferenceId)
        escrowAccount = await program.account["EscrowAccount"].fetch(escrowPDA)
        return escrowAccount


async def getModel(config, modelId: int) -> InvokerModel:
    async with getProgram(config) as program:
        modelPDA = getModelPDA(modelId)
        modelAccount = await program.account["InvokerModel"].fetch(modelPDA)
        return modelAccount


def passvar(x):
    pass


def get_or_create_token_account(
    client: Client, owner: Pubkey, tokenmint: Pubkey, payer: Keypair
) -> AccountInfo:
    addr = get_associated_token_address(owner, tokenmint)
    token = Token(client, tokenmint, TOKEN_PROGRAM_ID, payer)
    try:
        return token.get_account_info(addr)
    except Exception as e:
        passvar(e)
        token.create_associated_token_account(owner)
        return get_or_create_token_account(client, owner, tokenmint, payer)


async def test_pda(config):
    async with getProgram(config) as program:
        pk = Pubkey.from_string("F48tYLEGkyjt4PU5AFbe43fJcaQmG5AUh79GMBvFhUD1")
        escrowAccount = await program.account["EscrowAccount"].fetch(pk)
        return escrowAccount


async def descrow(config, modelid: int, inferenceid: int, output: str):
    wallet = getWallet(config)
    async with getProgram(config) as program:
        marketSettingsPDA, marketSettingsBump = getMarketSettingsPDA()
        marketSettings = await program.account["MarketSettings"].fetch(
            marketSettingsPDA
        )
        modelPDA = getModelPDA(modelid)
        model = await getModel(config, modelid)
        escrowAccountPDA = getEscrowAccountPDA(modelid, inferenceid)
        escrowWalletPDA = getEscrowWalletPDA(modelid, inferenceid)
        invokerTokenAccount = get_associated_token_address(
            wallet.public_key, model.token_mint
        )
        ownerTokenAccount = get_associated_token_address(
            marketSettings.owner, model.token_mint
        )
        accounts = {
            "invoker": wallet.public_key,
            "mint_of_token_being_sent": model.token_mint,
            "invoker_taker_token_account": invokerTokenAccount,
            "owner_token_account": ownerTokenAccount,
            "market_settings": marketSettingsPDA,
            "market_owner": marketSettings.owner,
            "model": modelPDA,
            "escrow_account": escrowAccountPDA,
            "escrow_wallet": escrowWalletPDA,
            "system_program": SYS_PROGRAM_ID,
            "token_program": TOKEN_PROGRAM_ID,
        }
        await program.rpc["descrow"](
            modelid,
            inferenceid,
            output,
            ctx=Context(accounts=accounts, signers=[getKeypair(config)]),
        )


async def addModel(
    config, modelDescription: str, tokenMint: Pubkey, price: int
) -> int:
    keypair = getKeypair(config)
    print("preparing deployment...")
    async with getProgram(config) as program:
        marketSettingsPDA, marketSettingsBump = getMarketSettingsPDA()
        marketSettings = await program.account["MarketSettings"].fetch(
            marketSettingsPDA
        )
        modelPDA = getModelPDA(marketSettings.model_id)
        accounts = {
            "market_owner": marketSettings.owner,
            "market_settings": marketSettingsPDA,
            "model": modelPDA,
            "user": keypair.pubkey(),
            "system_program": SYS_PROGRAM_ID,
        }
        await program.rpc["addmodel"](
            marketSettingsBump,
            modelDescription,
            price,
            tokenMint,
            ctx=Context(accounts=accounts, signers=[keypair]),
        )
        print("Transaction has been sent")
        return marketSettings.model_id



import logging
import os
from multiprocessing import Process
from time import sleep

from ..client import get_client
from ..server import task


def init_daemon(args, config):
    if os.fork() != 0:
        return
    process = Process(target=task, args=(args, config), daemon=True)
    # start the new process
    process.start()
    process.join()


def cmd_daemon(args, config):
    logging.info("Im the daemon")
    host = config["system"]["xml_rpc_addr"]
    port = config["system"]["xml_rpc_port"]
    c = get_client(host, port)
    if c is not None and c.isAlive() is True:
        logging.info("shutting down")
        c.quit()
    else:
        logging.info("spawning the daemon")
        init_daemon(args, config)
        sleep(2)



from typing import Literal

local_test_token_addr = "FzZDZZW9r6zg4Qnod6yxrDBkEghBdzmG59pL3SHzWhar"

dev_test_token_addr = "FzZDZZW9r6zg4Qnod6yxrDBkEghBdzmG59pL3SHzWhar"

mainnet_usdc_addr = "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v"
mainnet_tether_addr = "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB"

InvokerEnvironment = Literal["local", "dev", "mainnet"]


class InvokerEnv:
    LOCAL = "l"
    DEV = "d"
    MAINNET = "m"


TokenMap = {
    InvokerEnv.LOCAL: [
        {
            "TokenName": "InvokerLocalTestToken",
            "Address": local_test_token_addr,
        },
    ],
    InvokerEnv.DEV: [
        {"TokenName": "InvokerDevToken", "Address": dev_test_token_addr}
    ],
    InvokerEnv.MAINNET: [
        {"TokenName": "USDC (Circle)", "Address": mainnet_usdc_addr},
        {"TokenName": "USDT (Tether)", "Address": mainnet_tether_addr},
    ],
}



from .constants import InvokerEnv, InvokerEnvironment, TokenMap
from .storage.deployed_models import DeployedModel


def getTaskId(m: DeployedModel) -> str:
    return "{}_{}".format(m.env, m.model_id)


def IN_assert(input, output):
    assert input == output


def getEnvName(args) -> InvokerEnvironment:
    if args.env == InvokerEnv.LOCAL:
        return "local"
    if args.env == InvokerEnv.DEV:
        return "dev"
    if args.env == InvokerEnv.MAINNET:
        return "mainnet"
    raise ValueError("Invalid Environment")


def generateModelUrl(args, modelId) -> str:
    if args.env == InvokerEnv.LOCAL:
        return "http://localhost:3000/{}".format(modelId)
    if args.env == InvokerEnv.DEV:
        return "https://dev.invoker.network/{}".format(modelId)
    if args.env == InvokerEnv.MAINNET:
        return "https://invoker.network/{}".format(modelId)
    raise ValueError("Invalid Environment")


def getAvailableTokens(args):
    t = TokenMap.get(args.env, None)
    if t is None:
        raise ValueError("Invalid Environment")
    return t



import logging
from typing import get_args

from solana.rpc.api import Client
from solders.rpc.responses import RPCError, RPCResult

from ..constants import dev_test_token_addr
from ..cores.anchor import get_or_create_token_account
from ..cores.wallet import getPubkey, loadwallet

AIRDROP_AMOUNT = 10_000_000_000

RPC_RESULT_TYPES = get_args(RPCResult)


def assert_valid_response(resp: RPCResult):
    """Assert valid RPCResult."""
    assert type(resp) in RPC_RESULT_TYPES
    assert not isinstance(resp, RPCError.__args__)  # type: ignore


def cmd_wallet(args, config):
    logging.info("Wallet is loading...")
    keystr = config["system"]["wallet"]
    keypair = loadwallet(keystr)
    pubkey = keypair.pubkey()
    dev_client = Client(config["active"]["rpc_url"])
    # main_client = Client(config["active"]["rpc_url"])
    # tx = None
    # try:
    # tx = dev_client.request_airdrop(pubkey, 1000000000, "finalized")
    # assert_valid_response(tx)
    # dev_client.confirm_transaction(tx.value)
    # pass
    # except:
    # pass
    print("Wallet: {}".format(pubkey))

    # dev stats
    dev_balance = dev_client.get_balance(pubkey)
    print("Dev Balance: {}".format(dev_balance.value))
    dev_test_token_pub = getPubkey(dev_test_token_addr)
    dev_tether_balance = get_or_create_token_account(
        dev_client, pubkey, dev_test_token_pub, keypair
    )
    print("Dev Tether Balance: {}".format(dev_tether_balance.amount))

    """
    mainnet_balance = main_client.get_balance(pubkey).value
    print("*Mainnet Balance: {}".format(mainnet_balance))
    if mainnet_balance == 0:
        print("You need to deposit some coins to do mainnet operations.
        Please refer to our documentation. https://docs.invoker.network")
    else:
        mainnet_tether_token_pub = getPubkey(mainnet_tether_addr)
        mainnet_tether_balance = get_or_create_token_account(main_client,
        pubkey, mainnet_tether_token_pub, keypair)
        print("*Mainnet Tether Balance: {}".format(mainnet_
        tether_balance.amount))
    """

    # addr = get_associated_token_address(pubkey, dev_tether_pub)
    # print(addr)
    # print(dev_tether_token.get_account_info(addr))
    # print(dev_tether_token.get_balance(addr))

    # print(dev_tether_token.get_account_info(getPubkey("4YYKD5WMmQYyMXX7i7KpFfKHiDmKvpo16PtFVrcBiERE")))
    # print(dev_tether_token.get_balance(getPubkey("4YYKD5WMmQYyMXX7i7KpFfKHiDmKvpo16PtFVrcBiERE")))
    # print(dev_tether_token)
    # dev_tether_balance = dev_client.get_token_account_balance(dev_tether_pub)
    # print("Dev Tether Balance: {}".format(dev_tether_balance))
    # dev_usdc_pub = getPubkey(dev_usdc)
    # dev_usdc_balance = dev_client.get_token_account_balance(dev_usdc_pub)
    # print("Dev USDC Balance: {}".format(dev_usdc_balance))

    # mainnet stats
    # mainnet_balance = main_client.get_balance(pubkey)
    # print("Mainnet Balance: {}".format(mainnet_balance.value))
    # mainnet_tether_token = Token()
    # mainnet_tether_pub = getPubkey(main_tether)
    # mainnet_tether_token = Token(main_client, mainnet_tether_pub,
    # TOKEN_PROGRAM_ID, keypair)
    # acc = mainnet_tether_token.get_accounts_by_owner(pubkey)
    # acc = mainnet_tether_token.get_account_info(keypair.pubkey())
    # print(acc)
    # acc = mainnet_tether_token.create_account(pubkey)
    # mainnet_tether_token.get_account_info
    # mainnet_tether_account = mainnet_tether_token.get_balance(acc)
    # print("Mainnet Tether Balance: {}".format(mainnet_tether
    # _account.value.amount))
    # Token.create_account(mainnet_tether_token, pubkey)

    # print(mainnet_tether_pub)
    # opts = TokenAccountOpts(mint=mainnet_tether_pub)
    # mainnet_tether_account = main_client.get_token_accounts_
    # by_owner_json_parsed(pubkey, opts, 'finalized')
    # test = main_client.get_token_account_balance(mainnet_tether_pub)
    # print(mainnet_tether_account)
    # print(test)
    # mainnet_tether_balance = main_client.get_token_account_
    # balance(mainnet_tether_account)
    # print("Mainnet Tether Balance: {}".format(mainnet_tether_balance))
    # mainnet_usdc_pub = getPubkey(main_usdc)
    # mainnet_usdc_balance = main_client.get_token_account_
    # balance(mainnet_usdc_pub)
    # print("Mainnet USDC Balance: {}".format(mainnet_usdc_balance))



from .inputs.fileinput import InvokerNetworkFileInput
from .inputs.inputbase import InvokerNetworkInputBase
from .inputs.numberinput import InvokerNetworkNumberInput
from .inputs.rangeinput import InvokerNetworkRangeInput
from .inputs.selectinput import InvokerNetworkSelectInput
from .inputs.textinput import InvokerNetworkTextInput
from .outputs.output import InvokerOutput
from .inputs.decorators import invoker, invoker_input


from abc import ABC, abstractmethod


class InvokerValidationError(Exception):
    def __init__(self, message, errors):
        super().__init__(message)
        self.errors = errors


class InvokerField(ABC):
    name: str
    description: str

    @abstractmethod
    def validate(self):
        if self.name is None or len(self.name) == 0:
            raise InvokerValidationError("Name must exist")


class InvokerInputField(InvokerField):
    pass


def InvokerNetworkOutputBuilder():
    return None



"""CLI interface for invoker_terminal project.

Be creative! do whatever you want!

- Install click or typer and create a CLI app
- Use builtin argparse
- Start a web application
- Import things from your .base module
"""
import argparse
import configparser
import logging
import os
from argparse import RawTextHelpFormatter
from logging.handlers import RotatingFileHandler

from .commands.build import cmd_build
from .commands.daemon import cmd_daemon
from .commands.deploy import cmd_deploy
from .commands.history import cmd_history
from .commands.init import cmd_init
from .commands.invoke import cmd_invoke
from .commands.ipfs import cmd_ipfs
from .commands.new_task import cmd_new_task
from .commands.status import cmd_status, getEnvText
from .commands.test import cmd_test
from .commands.wallet import cmd_wallet
from .storage.helper import get_engine, init_tables

commandMap = {
    "init": cmd_init,
    "status": cmd_status,
    "daemon": cmd_daemon,
    "wallet": cmd_wallet,
    "new": cmd_new_task,
    "invoke": cmd_invoke,
    "build": cmd_build,
    "test": cmd_test,
    "deploy": cmd_deploy,
    "history": cmd_history,
    "ipfs": cmd_ipfs,
}

PROGRAM_NAME = "invoker_terminal"


def main():
    version = open(os.path.dirname(__file__) + "/VERSION").read()
    parser = argparse.ArgumentParser(
        prog=PROGRAM_NAME,
        description="""
        A terminal that connects to Invoker Network (https://invoker.network)
        """,
        formatter_class=RawTextHelpFormatter,
        epilog="""
    init: initialize a config and cache
    status: status information about daemon and dependencies
    daemon: start/stop daemon a.k.a the background process
    wallet: information about wallet & tokens
    new: generates a boilerplate for AI task
    build: builds the task
    invoke: invoke the  model (inside container)
    test: tests the task
    deploy: deploys the task
    history: shows recent transactions

    Version: {version}
    """.format(
            version=version
        ),
    )

    invo_config_dir = os.path.expanduser("~") + os.sep + ".invokernet"
    os.makedirs(invo_config_dir, exist_ok=True)
    invo_config = invo_config_dir + os.sep + "config.ini"
    db_path = invo_config_dir + os.sep + "invoker.db"
    log_path = invo_config_dir + os.sep + "system.log"

    config = configparser.ConfigParser()

    parser.add_argument(
        "command", nargs=1, choices=commandMap.keys(), default="status"
    )
    parser.add_argument(
        "--debug", action="store_true", help="enable debug environment"
    )
    parser.add_argument("--conf", action="append", default=[invo_config])

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--local",
        dest="env",
        action="store_const",
        const="l",
        help="local environment",
    )
    group.add_argument(
        "--dev",
        dest="env",
        action="store_const",
        const="d",
        help="dev environment",
    )
    group.add_argument(
        "--main",
        dest="env",
        action="store_const",
        const="m",
        help="prod environment",
    )
    parser.set_defaults(env="l", db_path=db_path, log_path=log_path)
    args = parser.parse_args()

    print("Config loaded from {}".format(args.conf))

    if args.conf is not None:
        for conf_fname in args.conf:
            if not os.path.isfile(conf_fname):
                logging.debug("Skipping {}. Not found!".format(conf_fname))
            else:
                logging.info("loading config from {}".format(conf_fname))
                config.read(conf_fname)

    logrotate = RotatingFileHandler(
        filename=log_path,
        mode="a",
        maxBytes=5 * 1024 * 1024,
        backupCount=1,
        encoding=None,
        delay=0,
    )

    if "system" in config:
        logging.basicConfig(
            format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
            level=int(config["system"]["log_level"]),
            handlers=[logrotate],
        )

    engine = get_engine(db_path)
    init_tables(engine)
    logging.info("Invoker DB loaded from {}".format(db_path))
    
    if not config.has_section("active"):
        config.add_section("active")

    if args.env == "l" and "local" in config:
        print("Environment " + getEnvText("LOCAL"))
        config["active"]["rpc_url"] = config["local"]["rpc_url"]
        config["active"]["websocket_url"] = config["local"]["websocket_url"]

    if args.env == "d" and "dev" in config:
        print("Environment " + getEnvText("DEV"))
        config["active"]["rpc_url"] = config["dev"]["rpc_url"]
        config["active"]["websocket_url"] = config["dev"]["websocket_url"]

    if args.env == "m" and "mainnet" in config:
        print("Environment " + getEnvText("MAINNET"))
        config["active"]["rpc_url"] = config["mainnet"]["rpc_url"]
        config["active"]["websocket_url"] = config["mainnet"]["websocket_url"]

    # print(config.has_section('system'))
    configLessCmds = ['init', 'invoke']
    cmd = args.command[0]
    if not cmd in configLessCmds:
        if not config.has_section("system"):
            parser.error(
                """

            needs a valid config before running

            type "{PROGRAM_NAME} init" to generate config

            """.format(
                    PROGRAM_NAME=PROGRAM_NAME
                )
            )

    # if args.command != ['deploy'] and args.prod:
    #    parser.error('--prod can only be set when deploy.')

    # if args.command == ['deploy'] and args.prod:
    #    if not config.has_section('prod'):
    #        parser.error('config needs to have --prod section')

    command = args.command[0]
    commandMap[command](args, config)



import io
import json
import os
import uuid
import zipfile
from pathlib import Path

import docker

from ..cores.wallet import generatewallet


def cmd_test(args, config):
    id = uuid.uuid4()
    host_url = os.path.join(config["system"]["tmp_folder"], str(id))
    print(host_url)
    Path(host_url).mkdir(parents=True, exist_ok=True)
    print(os.listdir(host_url))

    dummyinput = """
    {
    "description": "My own description",
    "inputs": [
        {
            "description": "count variable used in the model",
            "maxNum": 1000,
            "minNum": 100,
            "name": "count",
            "slider": false,
            "type": "number",
            "value": 500
        },
        {
            "description": "my description here",
            "maxchar": 200,
            "minchar": 10,
            "name": "prompt",
            "textarea": false,
            "type": "text",
            "value": "a generous prompt is here"
        },
        {
            "description": "",
            "maxchar": -1,
            "minchar": -1,
            "name": "prompt2",
            "textarea": false,
            "type": "text",
            "value": "a new version of promppt2 is here"
        }
    ],
    "name": "Facebook llama model"
}

"""
    zip_buffer = io.BytesIO()
    with zipfile.ZipFile(
        zip_buffer, "a", zipfile.ZIP_DEFLATED, False
    ) as zip_file:
        for file_name, data in [
            ("input.json", io.BytesIO(dummyinput.encode("utf-8"))),
            ("2.txt", io.BytesIO(b"222")),
        ]:
            zip_file.writestr(file_name, data.getvalue())

    with open("{}/input.zip".format(host_url), "wb") as f:
        f.write(zip_buffer.getvalue())

    client = docker.DockerClient(base_url=config["system"]["docker_url"])
    volumes = [host_url]
    volume_bindings = {
        "{}".format(host_url): {
            "bind": "/invoker",
            "mode": "rw",
        },
    }

    host_config = client.api.create_host_config(binds=volume_bindings)

    f = json.load(open("./build.json"))
    print("id is {}".format(f["id"]))
    container = client.api.create_container(
        f["id"], volumes=volumes, host_config=host_config
    )
    response = client.api.start(container=container.get("Id"))
    logs = client.api.logs(container=container.get("Id"))
    print("--logs--")
    print(logs)
    print(response)
    print(container)
    print(type(container))
    # print(container.logs())

    print(os.listdir(host_url))

    wallet = generatewallet()
    print(wallet)
    print(wallet.to_json())
    # delete the temp folder
    try:
        # shutil.rmtree(host_url)
        # print("DELETE - OK")
        pass
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))



import asyncio
import io
import json
import logging
import os
import time
import zipfile

from pick import pick
from solders.pubkey import Pubkey

from ..cores.anchor import addModel
from ..cores.ipfs import send_to_ipfs
from ..storage.deployed_models import addDeployedModel
from ..storage.helper import get_engine
from ..utils import generateModelUrl, getAvailableTokens, getEnvName


def cmd_deploy(args, conf):
    print("deploy" + str(args))
    isLLM = False
    option, index = pick(["Yes", "No"], "Do you want LLM layout ?")
    if index == 0:
        isLLM = True
    availableTokens = getAvailableTokens(args)
    selectedToken = availableTokens[0]["TokenName"]
    selectedTokenAddr = availableTokens[0]["Address"]
    if len(availableTokens) > 1:
        print("handle the option here")
        tokenNames = list(map(lambda x: x["TokenName"], availableTokens))
        print(tokenNames)
        o, i = pick(tokenNames, "Which token do you want to use ?")
        selectedToken = availableTokens[i]["TokenName"]
        selectedTokenAddr = availableTokens[i]["Address"]
    print("SelectedToken {}".format(selectedToken))
    print("Selected Addr {}".format(selectedTokenAddr))
    price = int(input("Enter model price:"))
    ipfs_post_url = conf["system"]["ipfs_post_gateway"]
    filesMustExist = ["build.json", "desc.json"]
    for file in filesMustExist:
        if not os.path.isfile(file):
            logging.info(f"I can't deploy, mising file {file}")
            os._exit(1)
    descfile = open("desc.json", "r").read()
    buildfile = json.loads(open("build.json", "r").read())
    if isLLM:
        descfile = json.loads(descfile)
        descfile["layout"] = "LLM"
        descfile = json.dumps(descfile)
    zip_buffer = io.BytesIO()
    zip_file = zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False)
    zip_file.writestr("model/desc.json", descfile)
    ts = time.time()
    zip_file.writestr("timestamp", str(ts))
    zip_file.close()
    buff = zip_buffer.getvalue()
    modelDescHash = send_to_ipfs(buff, ipfs_post_url)
    # load anchor now
    modelId = asyncio.run(
        addModel(
            conf, modelDescHash, Pubkey.from_string(selectedTokenAddr), price
        )
    )
    print("Check out the model at {}".format(generateModelUrl(args, modelId)))
    imageId = buildfile["id"]
    modelname = os.path.split(os.getcwd())[1]
    print("Current model name {}".format(modelname))
    addDeployedModel(
        get_engine(args.db_path), getEnvName(args), modelId, imageId, modelname
    )



from .inputbase import InvokerNetworkInputBase


class InvokerNetworkNumberInput(InvokerNetworkInputBase):
    def __init__(
        self,
        description: str = None,
        required: bool = False,
        minNum: int = None,
        maxNum: int = None,
    ):
        super().__init__(description, required)
        if minNum is not None:
            if not isinstance(minNum, int):
                raise ValueError(
                    "Invalid value for minNum expected \
                        int got {} instead".format(
                        minNum
                    )
                )
            else:
                self.minnum = minNum
        if maxNum is not None:
            if not isinstance(maxNum, int):
                raise ValueError(
                    "Invalid value for maxNum expected \
                        int got {} instead".format(
                        maxNum
                    )
                )
            else:
                self.maxnum = maxNum
        self.type = "number"



import json
import zipfile


def Invoker_init():
    print("invoker init here for containers")
    pass


class dummyObj(object):
    pass


# parse the input.zip file
# then populate the object accordingly
def ParseInput(path=None):
    if path is None:
        path = "/invoker/input.zip"
    archive = zipfile.ZipFile(path, "r")
    inputdata = archive.read("input/desc.json").decode("utf-8")
    dummy = dummyObj()
    parsed = json.loads(inputdata)
    # validate the input here
    for key in parsed:
        val = parsed[key]
        if val == "@@linked":
            file = archive.read("input/files/{}".format(key))
            setattr(dummy, key, file)
        else:
            setattr(dummy, key, val)
    return dummy



import logging
import os

DOCKERFILE_FILE = (
    os.path.dirname(__file__) + "/../templates/Dockerfile_template"
)
MODEL_FILE = os.path.dirname(__file__) + "/../templates/model.py"

model_template = open(MODEL_FILE).read()
dockerfile_template = open(DOCKERFILE_FILE).read()


def cmd_new_task(args, config):
    cwd = os.getcwd()
    name = str(input("Enter model name: "))
    print(name)
    print(cwd)
    if os.path.isdir(name):
        logging.info("Directory exists {name}".format(name=name))
        os._exit(1)
    elif not os.access(cwd, os.W_OK):
        logging.info("Directory is not writable... {dir}".format(dir=cwd))
        os._exit(1)
    else:
        logging.info("Creating task files...")
        newpath = cwd + "/" + name
        os.makedirs(newpath, exist_ok=False)

        # dockerfile
        f = open(newpath + "/Dockerfile", "w+")
        f.write(dockerfile_template)
        f.close()

        # modelpy
        f = open(newpath + "/model.py", "w")
        f.write(model_template)
        f.close()

    # name = input('What is the task name? (must be unique)\n')
    # print(name)



from solders.keypair import Keypair
from solders.pubkey import Pubkey


def generatewallet() -> Keypair:
    return Keypair()


def loadwallet(js) -> Keypair:
    kp = Keypair.from_json(js)
    return kp


def getPubkey(pk: str) -> Pubkey:
    return Pubkey.from_string(pk)



import asyncio
import sys

import docker
from termcolor import colored

from ..client import get_client
from ..cores.ipfs import status_ipfs


def getEnvText(name):
    return colored("[ {} ]".format(name), "green", attrs=["reverse"])


def getFailText():
    return colored("[ FAIL ]", "red", attrs=["reverse"])


def getSuccessText():
    return colored("[ OK ]", "green", attrs=["reverse"])


async def test_run(cf):
    # escrowAccount = await getEscrow(cf, 288, 0)
    # print(escrowAccount)
    # testoutput = "QmT7ynzTEFK8DVEM48RfVo4Rumy4qyMi9Y9doh3doDyJbZ"
    # await descrow(cf, 288, 0, testoutput)
    # a = await test_pda(cf)
    # print(a)
    pass


# returns output


def cmd_status(args, config):
    print("Version " + sys.version)
    print("Version Info " + str(sys.version_info))
    text = ""
    # docker connection
    try:
        docker.DockerClient(base_url=config["system"]["docker_url"])
        text = getSuccessText()
    except Exception as e:
        print(e)
        text = getFailText()
    print("Docker " + text)
    # daemon connection
    host = config["system"]["xml_rpc_addr"]
    port = config["system"]["xml_rpc_port"]
    c = get_client(host, port)
    if c is None or c.isAlive() is not True:
        text = getFailText()
    else:
        text = getSuccessText()
    print("Daemon " + text)

    ipfs = status_ipfs(
        config["system"]["ipfs_get_gateway"]
        + "/"
        + config["system"]["ipfs_test_hash"]
    )
    if ipfs:
        text = getSuccessText()
    else:
        text = getFailText()
    print("IPFS " + text)

    if c:
        models = c.get_contracts_list()
        print("models")
        print(models)

    asyncio.run(test_run(config))
